import code
from tokenize import group
from ultron.factor.experimental.normalize import rolling_groups
from ultron.factor.analysis.quantile_analysis import er_quantile_analysis
from ultron.factor.data.quantile import quantile
from ultron.tradingday import *
import numpy as np
import pandas as pd
import pdb, copy
class SingleFactor(object):
    def __init__(self, factor_data, market_data, codes, columns):
        self._factor_data = factor_data
        self._market_data = market_data
        self._codes = codes
        self._columns = columns

    def returns(self, market_data, period):
        price_tb = market_data['closePrice'].unstack()
        price_tb.fillna(method='pad', inplace=True)
        return_tb = np.log(price_tb.shift(-period) / price_tb)
        return_tb = return_tb.replace([np.inf, -np.inf], np.nan)
        return_tb = return_tb.stack().reindex(market_data.index)
        return_tb.name = 'nxt1_ret'
        return return_tb.reset_index()

    def normalize(self, factor_data, windows, columns):
        #normalize_data = factor_data.set_index('code').groupby(level=['code']).apply(
        #        lambda x: rolling_groups(x,columns, windows))
        #normalize_data = normalize_data.reset_index().drop(['level_1'],axis=1)
        rolling_data = factor_data.set_index(['trade_date','code']).unstack().rolling(windows)
        current_data = factor_data.set_index(['trade_date','code']).unstack()
        normalize_data = (current_data - rolling_data.mean()) / rolling_data.std()
        normalize_data = normalize_data.stack(dropna=True)
        normalize_data = normalize_data.sort_values(by=['trade_date','code']).fillna(0)
        return normalize_data.reset_index()

    
    def _transformer(self, normalize_data, returns, columns, period):
        begin_date = normalize_data.trade_date.min()
        end_date = normalize_data.trade_date.max()
        dates = makeSchedule(begin_date, end_date,  str(period)+'b', 'china.sse', 
                            BizDayConventions.Preceding)
        dates = [d.strftime('%Y-%m-%d') for d in dates]
        dt = normalize_data.trade_date.dt.strftime('%Y-%m-%d').unique().tolist()
        dates = set(dt) & set(dates)
        normalize_data = normalize_data.set_index('trade_date').loc[dates].sort_values(
            by=['trade_date']).reset_index()
        normalize_data['trade_date'] = pd.to_datetime(normalize_data['trade_date'])
        returns['trade_date'] = pd.to_datetime(returns['trade_date'])
        total_data = returns.merge(normalize_data, on=['trade_date','code'], how='left')
        total_data = total_data.set_index(
            ['trade_date','code'])[columns].unstack().fillna(
                method='pad').stack().reset_index()
        return total_data

    def quantile_analysis(self, normalize_data, factor_name, n_bins, de_trend=False):
        df = pd.DataFrame(columns=['q' + str(i) for i in range(1, n_bins+1)])
        grouped = normalize_data.groupby('trade_date')
        for k, g in grouped:
            er = g[factor_name].values
            dx_return = g['nxt1_ret'].values
            res = er_quantile_analysis(er, n_bins=n_bins, dx_return=dx_return, de_trend=de_trend)
            df.loc[k, :] = res
        df['q'] = df['q'+str(n_bins)]-df['q1']
        return df
    
    def quantile(self, normalize_data, factor_name, n_bins):
        grouped = normalize_data.groupby('trade_date')
        res = []
        for k, g in grouped:
            o = copy.deepcopy(g)
            t = copy.deepcopy(o)
            o = o.dropna(subset=[factor_name])
            o['group'] = quantile(o[factor_name], n_bins) + 1
            o = t.merge(o[['code','trade_date','group']], on=['trade_date','code'], how='left')
            res.append(o.set_index(['trade_date','code']))
        group_dt = pd.concat(res,axis=0)
        return group_dt


    def run(self, codes=None, columns=[], windows=10, period=1, n_bins=5, 
            normalize_data=None, returns_data=None):
        print('start single factor analysis...')
        if len(columns) == 0:
            columns = self._columns
        if normalize_data is None:
            print('start data normalize...')
            normalize_data = self.normalize(factor_data=self._factor_data, windows=windows, 
                            columns=columns)
        normalize_data = self._factor_data
        if returns_data is None:
            print('start data returns...')
            returns_data = self.returns(self._market_data.set_index(['trade_date','code']), 
                            period)
        if codes is None:
            codes = self._codes
            
        normalize_data['trade_date'] = pd.to_datetime(normalize_data['trade_date'])
        returns_data['trade_date'] = pd.to_datetime(returns_data['trade_date'])
        total_data = normalize_data.merge(returns_data, on=['trade_date','code'])
        
        ### 指定品种
        codes = list(set(total_data['code'].unique().tolist()) & set(codes))
        total_data = total_data.set_index('code').loc[codes].reset_index()
        qt_res = []
        for col in columns:
            print("start {0} quantile ...".format(col))
            df = self.quantile(normalize_data[['code','trade_date', col]], col, n_bins)
            qt_res.append({'name':col, "qdata":df})

        qa_res = []
        for col in columns:
            print("start {0} quantile analysis...".format(col))
            df = self.quantile_analysis(normalize_data=total_data, 
                    factor_name=col, n_bins=n_bins)
            qa_res.append({'name':col, "qdata":df})
        return qt_res, qa_res
        




# -*- coding: utf-8 -*-

import os,os.path
import zipfile

def zip_compress(dir_name, zip_filename):
    filelist = []
    if os.path.isfile(dir_name):
        filelist.append(dir_name)
    else :
        for root, dirs, files in os.walk(dir_name):
            for name in files:
                filelist.append(os.path.join(root, name))
          
    zf = zipfile.ZipFile(zip_filename, "w", zipfile.zlib.DEFLATED)
    for tar in filelist:
        arcname = tar[len(dir_name):]
        zf.write(tar,arcname)
    zf.close()

def unzip_compress(zip_filename, unzip_dir):
    if not os.path.exists(unzip_dir): 
        os.mkdir(unzip_dir)
    zfobj = zipfile.ZipFile(zip_filename)
    for name in zfobj.namelist():
        name = name.replace('\\','/')
        if name.endswith('/'):
            os.mkdir(os.path.join(unzip_dir, name))
        else:           
            ext_filename = os.path.join(unzip_dir, name)
            ext_dir= os.path.dirname(ext_filename)
            if not os.path.exists(ext_dir) : 
                os.mkdir(ext_dir)
            outfile = open(ext_filename, 'wb')
            outfile.write(zfobj.read(name))
            outfile.close()


# -*- coding: utf-8 -*-
u"""
Created on 2017-6-11

@author: cheng.li
"""

from ultron.tradingday.Period import Period
from ultron.tradingday.Enums.TimeUnits import TimeUnits
from ultron.tradingday.Enums.Months import Months


class DayCounter(object):

    def __init__(self, dcName):

        dcName = dcName.lower()
        if dcName in _dcDict:
            self.impl_ = _dcDict[dcName]()
        else:
            raise ValueError('{0} is not a valid day counter name'.format(dcName))

    def name(self):
        return self.impl_.name()

    def dayCount(self, d1, d2):
        return self.impl_.dayCount(d1, d2)

    def yearFraction(self, d1, d2, refPeriodStart=None, refPeriodEnd=None):
        return self.impl_.yearFraction(d1, d2, refPeriodStart, refPeriodEnd)


class DayCounterImpl(object):

    def name(self):
        pass

    def dayCount(self, d1, d2):
        return d2 - d1

    def yearFraction(self, d1, d2, refPeriodStart, refPeriodEnd):
        pass


class Actual360(DayCounterImpl):

    def name(self):
        return 'Actual/360'

    def yearFraction(self, d1, d2, refPeriodStart, refPeriodEnd2):
        return float(d2 - d1) / 360.


class Actual365Fixed(DayCounterImpl):

    def name(self):
        return 'Actual/365 (Fixed)'

    def yearFraction(self, d1, d2, refPeriodStart, refPeriodEnd2):
        return float(d2 - d1) / 365.


class Actual365NoLeap(DayCounterImpl):
    MonthOffset = [
        0, 31, 59, 90, 120, 151,  # Jan - Jun
        181, 212, 243, 273, 304, 334  # Jun - Dec
    ]

    def name(self):
        return 'Actual/365 (NL)'

    def dayCount(self, d1, d2):

        s1 = d1.dayOfMonth() + self.MonthOffset[d1.month() - 1] + (d1.year() * 365)
        s2 = d2.dayOfMonth() + self.MonthOffset[d2.month() - 1] + (d2.year() * 365)

        if d1.month() == Months.Feb and d1.dayOfMonth() == 29:
            s1 -= 1

        if d2.month() == Months.Feb and d2.dayOfMonth() == 29:
            s2 -= 1

        return s2 - s1

    def yearFraction(self, d1, d2, refPeriodStart, refPeriodEnd2):
        return self.dayCount(d1, d2) / 365.


class ActualActualISMAImpl(DayCounterImpl):

    def name(self):
        return 'Actual/Actual (ISMA)'

    def yearFraction(self, d1, d2, d3, d4):

        if d1 == d2:
            return 0.

        if d1 > d2:
            return -self.yearFraction(d2, d1, d3, d4)

        refPeriodStart = d3 if d3 else d1
        refPeriodEnd = d4 if d4 else d2

        months = int(0.5 + 12. * (refPeriodEnd - refPeriodStart) / 365.)

        if months == 0:
            refPeriodStart = d1
            refPeriodEnd = d1 + '1y'
            months = 12

        period = months / 12.

        if d2 <= refPeriodEnd:
            if d1 >= refPeriodStart:
                return period * (d2 - d1) / (refPeriodEnd - refPeriodStart)
            else:
                previousRef = refPeriodStart - Period(length=months, units=TimeUnits.Months)

                if d2 > refPeriodStart:
                    return self.yearFraction(d1, refPeriodStart, previousRef, refPeriodStart) \
                           + self.yearFraction(refPeriodStart, d2, refPeriodStart, refPeriodEnd)
                else:
                    return self.yearFraction(d1, d2, previousRef, refPeriodStart)

        else:
            sum = self.yearFraction(d1, refPeriodEnd, refPeriodStart, refPeriodEnd)
            i = 0
            while True:
                newRefStart = refPeriodEnd + Period(length=i*months, units=TimeUnits.Months)
                newRefEnd = refPeriodEnd + Period(length=(i+1)*months, units=TimeUnits.Months)

                if d2 < newRefEnd:
                    break
                else:
                    sum += period
                    i += 1

            sum += self.yearFraction(newRefStart, d2, newRefStart, newRefEnd)
            return sum


_dcDict = {'actual/360': Actual360,
           'actual/365 (fixed)': Actual365Fixed,
           'actual/365 (nl)': Actual365NoLeap,
           'actual/actual (isma)': ActualActualISMAImpl}



import os,time, pdb
import datetime as dt
import sqlalchemy
import numpy as np
import pandas as pd
import statsmodels.api as sm
from . long_short import LongShortWeighted
from . only_side import OnlySideWeighted
from . high_frequency import HighFrequencyWeighted
from . basic_indicators import IC_Weighted

class Weighted(object):
    @classmethod
    def create_weighted(cls, method='longshort'):
        if method == 'longshort':
            return LongShortWeighted()
        elif method == 'onlyside':
            return OnlySideWeighted()
        elif method == 'high_frequency':
            return HighFrequencyWeighted()
        elif method == 'ic':
            return IC_Weighted('ic')
        elif method == 'ir':
            return IC_Weighted('ir')
        else:
            return None


import numpy as np
import pandas as pd
import pdb
from ultron.factor.data.standardize import standardize
from ultron.factor.data.processing import factor_processing

def rolling_standard(x, name, res):
    y = x.values.reshape(x.values.shape[0],1)
    values = factor_processing(y, pre_process=[standardize])
    values = pd.Series(values.reshape(x.values.shape[0],), x.index)
    values.name = name
    res.append(values.reset_index().to_dict(orient='records')[-1])
    return 1


def rolling_groups(data, columns, windows=10):
    alpha_res = []
    for col in columns:
        res = []
        _ = data.set_index('trade_date')[[col]].rolling(windows).apply(
            lambda x: rolling_standard(x, col, res))
        values = pd.DataFrame(res).set_index(['trade_date'])
        alpha_res.append(values)
    return pd.concat(alpha_res,axis=1).reset_index()


# -*- coding: utf-8 -*-
import six
from ultron.utilities.singleton import Singleton
from ultron.utilities.redis.redis_client import RedisClient
import ultron.proto.cache_data_pb2
from ultron.config import config_setting
import zlib
import base64
import pdb

@six.add_metaclass(Singleton)
class CacheData(object):
    def __init__(self, **kwargs):
        if ('host'  in kwargs) and ('port'  in kwargs) and ('pwd'  in kwargs):
            self._host = kwargs['host']
            self._port = kwargs['port']
            self._port = kwargs['pwd']
        else:
            self._host = config_setting.queue_host
            self._port = config_setting.queue_port
            self._pwd = config_setting.queue_pwd
        self._queue = 'ultron:cache'
        self._zip_level = 9
        #创建redis连接
        self._redis_client = RedisClient(host=self._host,
                                          port=self._port,
                                          password=self._pwd, db=6)
        
    def set_cache(self, session, key, values):
        cache_data_pb = ultron.proto.cache_data_pb2.CacheData()
        cache_data_pb.session = str(session).encode('raw_unicode_escape')
        cache_data_pb.key = str(key).encode('raw_unicode_escape')
        cache_data_pb.data_stream = zlib.compress(bytes(values,encoding='utf8'), self._zip_level)
        self._redis_client.hset(self._queue, str(key) + str(session), 
                                base64.b64encode(cache_data_pb.SerializeToString()))
    
    def get_cache(self, session, key, is_del=True):
        base64_values_str = self._redis_client.hget(self._queue, str(key) + str(session))
        if is_del:
            self._redis_client.hdel(self._queue, str(key) + str(session))
        zlilb_values = base64.b64decode(base64_values_str)
        cache_data_pb = ultron.proto.cache_data_pb2.CacheData()
        cache_data_pb.ParseFromString(zlilb_values)
        values = zlib.decompress(cache_data_pb.data_stream)
        return values
    
cache_data = CacheData()


# -*- coding: utf-8 -*-
import pdb
import datetime
import json
import subprocess
from ultron.cluster.work.extern_modules.base_module import BaseModule

class Module(BaseModule):
    def __init__(self, name, wid, token, redis_client):
        super(Module, self).__init__(name, wid, token, redis_client)
        self._func = {'startup_task':self.startup_task,
                     'shutoff_task':self.shutoff_task,
                     'restart_task':self.restart_task,
                     'update_task': self.update_task}
        self._namespace = 'tasks'
        
    def process_respone(self, respone):
        name = respone['name']
        opcode = respone['opcode']
        self._func[opcode](respone)
    
    def update_task(self, respone):
        dir_name = respone['dir_name']
        tasks = respone['tasks']
        for task in tasks:
            dir_line = 'cd ' + dir_name
            stop_line = 'ps auxww | grep \'active- (-A  '+ str(task) + ' worker\' | awk \'{print $2}\' | xargs kill -9'
            subprocess.Popen(stop_line, shell=True)
            celery_line = 'celery -A ' + task + ' worker --loglevel=info'
            command_line = dir_line + ' && ' + celery_line
            subprocess.Popen(command_line, shell=True)
        
    def startup_task(self, respone):
        work_name = respone['work_name']
        dir_name = respone['dir_name']
        #celery -A polymeriza worker --loglevel=info
        dir_line = 'cd ' + dir_name
        stop_line = 'ps auxww | grep \'active- (-A  '+ str(work_name) + ' worker\' | awk \'{print $2}\' | xargs kill -9'
        celery_line = 'celery -A ' + work_name + ' worker --loglevel=info -E'
        command_line = dir_line + ' && ' + celery_line
        subprocess.Popen(command_line, shell=True)
    
    def shutoff_task(self, respone):
        work_name = respone['work_name']
        dir_name = respone['dir_name']
        command_line = 'ps auxww | grep \'active- (-A  '+ str(work_name) + ' worker\' | awk \'{print $2}\' | xargs kill -9'
        subprocess.Popen(command_line, shell=True)
    
    def restart_task(self):
        work_name = respone['work_name']
        dir_name = respone['dir_name']
        dir_line = 'cd ' + dir_name
        kill_line = 'ps auxww | grep \'active- (-A  '+ str(work_name) + ' worker\' | awk \'{print $2}\' | xargs kill -9'
        start_line = 'celery -A ' + work_name + ' worker --loglevel=info'
        command_line = dir_line + ' && ' + kill_line + ' && ' + start_line
        subprocess.Popen(command_line, shell=True)
        

    


import numpy as np

from .pfopt.linear import LpOptimizer as _LpOptimizer
from .pfopt.linear import L1LpOptimizer as _L1LpOptimizer
from .pfopt.quadratic import QOptimizer as _QOptimizer
from .pfopt.quadratic import DecomposedQOptimizer as _DecomposedQOptimizer
from .pfopt.quadratic import TargetVarianceOptimizer as _TargetVarianceOptimizer



class LPOptimizer:

    def __init__(self,
                 objective: np.array,
                 cons_matrix: np.ndarray,
                 lbound: np.ndarray,
                 ubound: np.ndarray,
                 method: str = "deprecated"):
        self._optimizer = _LpOptimizer(cost=objective,
                                       cons_matrix=cons_matrix,
                                       lower_bound=lbound,
                                       upper_bound=ubound)
        self._x, self._f_eval, self._status = self._optimizer.solve(solver="ECOS")

    def status(self):
        return self._status

    def feval(self):
        return self._f_eval

    def x_value(self):
        return self._x


class L1LPOptimizer:

    def __init__(self,
                 objective: np.array,
                 cons_matrix: np.ndarray,
                 current_pos: np.ndarray,
                 target_turn_over: float,
                 lbound: np.ndarray,
                 ubound: np.ndarray):
        self._optimizer = _L1LpOptimizer(cost=objective,
                                         benchmark=current_pos,
                                         l1norm=target_turn_over,
                                         cons_matrix=cons_matrix,
                                         lower_bound=lbound,
                                         upper_bound=ubound)
        self._x, self._f_eval, self._status = self._optimizer.solve()

    def status(self):
        return self._status

    def feval(self):
        return self._f_eval

    def x_value(self):
        return self._x


class QuadraticOptimizer:

    def __init__(self,
                 objective: np.array,
                 cons_matrix: np.ndarray = None,
                 lbound: np.ndarray = None,
                 ubound: np.ndarray = None,
                 penalty: float = 1.,
                 cov: np.ndarray = None,
                 factor_cov: np.ndarray = None,
                 factor_load: np.ndarray = None,
                 factor_special: np.ndarray = None):
        if cov is None and factor_cov is not None:
            self._optimizer = _DecomposedQOptimizer(cost=objective,
                                                    factor_var=factor_cov,
                                                    factor_load=factor_load,
                                                    factor_special=factor_special,
                                                    penalty=penalty,
                                                    cons_matrix=cons_matrix,
                                                    lower_bound=lbound,
                                                    upper_bound=ubound)
        elif cov is not None:
            self._optimizer = _QOptimizer(cost=objective,
                                          variance=cov,
                                          penalty=penalty,
                                          cons_matrix=cons_matrix,
                                          lower_bound=lbound,
                                          upper_bound=ubound)
        else:
            raise ValueError("cov and factor cov can't be all empty")
        self._x, self._f_eval, self._status = self._optimizer.solve()

    def status(self):
        return self._status

    def feval(self):
        return self._f_eval

    def x_value(self):
        return self._x


class TargetVolOptimizer:

    def __init__(self,
                 objective: np.array,
                 cons_matrix: np.ndarray = None,
                 lbound: np.ndarray = None,
                 ubound: np.ndarray = None,
                 target_vol: float = 1.,
                 cov: np.ndarray = None,
                 factor_cov: np.ndarray = None,
                 factor_load: np.ndarray = None,
                 factor_special: np.ndarray = None):
        if cov is None and factor_cov is not None:
            self._optimizer = _TargetVarianceOptimizer(cost=objective,
                                                       variance_target=target_vol*target_vol,
                                                       factor_var=factor_cov,
                                                       factor_load=factor_load,
                                                       factor_special=factor_special,
                                                       cons_matrix=cons_matrix,
                                                       lower_bound=lbound,
                                                       upper_bound=ubound)
        elif cov is not None:
            self._optimizer = _TargetVarianceOptimizer(cost=objective,
                                                       variance_target=target_vol*target_vol,
                                                       variance=cov,
                                                       cons_matrix=cons_matrix,
                                                       lower_bound=lbound,
                                                       upper_bound=ubound)
        else:
            raise ValueError("cov and factor cov can't be all empty")
        self._x, self._f_eval, self._status = self._optimizer.solve()

    def status(self):
        return self._status

    def feval(self):
        return self._f_eval

    def x_value(self):
        return self._x


# -*- coding: utf-8 -*-

from ultron.tradingday.Date cimport Date
from ultron.tradingday.Calendar cimport Calendar
from ultron.tradingday.Period cimport Period


cdef class Schedule(object):

    cdef public _effectiveDate
    cdef public _terminationDate
    cdef public Period _tenor
    cdef public Calendar _cal
    cdef public int _convention
    cdef public int _terminationConvention
    cdef public int _rule
    cdef public list _dates
    cdef public list _isRegular
    cdef public bint _endOfMonth
    cdef public Date _firstDate
    cdef public Date _nextToLastDate

    cpdef size_t size(self)
    cpdef bint isRegular(self, size_t i)
    cpdef Calendar calendar(self)
    cpdef Period tenor(self)
    cpdef bint endOfMonth(self)


import numbers

import numpy as np
from joblib import cpu_count

def _get_n_jobs(n_jobs):
    """Get number of jobs for the computation.

    This function reimplements the logic of joblib to determine the actual
    number of jobs depending on the cpu count. If -1 all CPUs are used.
    If 1 is given, no parallel computing code is used at all, which is useful
    for debugging. For n_jobs below -1, (n_cpus + 1 + n_jobs) are used.
    Thus for n_jobs = -2, all CPUs but one are used.

    Parameters
    ----------
    n_jobs : int
        Number of jobs stated in joblib convention.

    Returns
    -------
    n_jobs : int
        The actual number of jobs as positive integer.

    """
    if n_jobs < 0:
        return max(cpu_count() + 1 + n_jobs, 1)
    elif n_jobs == 0:
        raise ValueError('Parameter n_jobs == 0 has no meaning.')
    else:
        return n_jobs


def partition_estimators(n_estimators, n_jobs):
    """Private function used to partition estimators between jobs."""
    # Compute the number of jobs
    n_jobs = min(_get_n_jobs(n_jobs), n_estimators)

    # Partition estimators between jobs
    n_estimators_per_job = (n_estimators // n_jobs) * np.ones(n_jobs,
                                                              dtype=np.int)
    n_estimators_per_job[:n_estimators % n_jobs] += 1
    starts = np.cumsum(n_estimators_per_job)

    return n_jobs, n_estimators_per_job.tolist(), [0] + starts.tolist()


import os,time, pdb
import datetime as dt
import sqlalchemy
import numpy as np
import pandas as pd
import statsmodels.api as sm

class LongShortWeighted(object):
    def __init__(self):
        self._industry_styles = ['Bank','RealEstate','Health','Transportation','Mining',
                                 'NonFerMetal','HouseApp','LeiService','MachiEquip','BuildDeco',
                                 'CommeTrade','CONMAT','Auto','Textile','FoodBever','Electronics',
                                 'Computer','LightIndus','Utilities','Telecom','AgriForest','CHEM',
                                 'Media','IronSteel','NonBankFinan','ELECEQP','AERODEF','Conglomerates']
        self._risk_styles = ['BETA','MOMENTUM','SIZE','EARNYILD','RESVOL','GROWTH','BTOP',
                             'LEVERAGE','LIQUIDTY','SIZENL']
        
    #暂时写此处，去极值
    def winsorize(self, total_data, method='sigma', limits=(3.0, 3.0), drop=True):
        se = total_data.copy()
        if method == 'quantile':
            down, up = se.quantile([limits[0], 1.0 - limits[1]])
        elif method == 'sigma':
            std, mean = se.std(), se.mean()
            down, up = mean - limits[0]*std, mean + limits[1]*std
        
        if drop:
            se[se<down] = np.NaN
            se[se>up] = np.NaN
        else:
            se[se<down] = down
            se[se>up] = up
        return se
    
    #标准化
    def standardize(self, total_data):
        try:
            res = (total_data - total_data.mean()) / total_data.std()
        except:
            res = pd.Series(data=np.NaN, index=total_data.index)
        return res
    
    #中性化
    def neutralize(self, total_data, risk_df):
        se = total_data.dropna()
        # se = total_data.copy()
        risk = risk_df.loc[se.index,:]
        # use numpy for neu, which is faster
        x = np.linalg.lstsq(risk.values, np.matrix(se).T)[0]
        se_neu = se - risk.dot(x)[0]
    
        return se_neu

    def _to_weights(self, group):
        demeaned_vals = group - group.mean()
        return demeaned_vals / demeaned_vals.abs().sum()
    
    
    def _to_ls_count(self, group, long=True):
        demeaned_vals = group - group.mean()
        if long:
            count = len(demeaned_vals[demeaned_vals>0])
        else:
            count = len(demeaned_vals[demeaned_vals<0])
        return count
    
    def returns(self, total_data, forward_returns, init_capital=100000):
        weights = total_data.groupby(level=['trade_date']).apply(self._to_weights)
        weighted_returns = forward_returns.multiply(weights, axis=0)
        
        factor_ret = weighted_returns.groupby(level='trade_date').sum()
        pnl = init_capital*2*factor_ret
        turnover = weights.unstack().T.diff().abs().sum(axis=1)
        long_count = total_data.groupby(level=['trade_date']).apply(self._to_ls_count)
        short_count = total_data.groupby(level=['trade_date']).apply(lambda x: self._to_ls_count(x, long=False))
        return factor_ret, pnl, turnover, long_count, short_count
    
    def run(self, factor_data, risk_data, forward_returns,
            factor_name, horizon=1, method='quantile', up_limit=0.025, down_limit=0.025,
            init_capital=100000):
        """
        参数：
            horizon: 调仓期，按照交易日计算。
        """
        factor_se = factor_data.set_index(['code','trade_date'])
        risk_se = risk_data.set_index(['code','trade_date'])
        forward_returns = forward_returns.set_index(['code','trade_date'])
        # risk_df = risk_se.reindex(factor_se.index)[self._industry_styles + ['SIZE'] + ['COUNTRY']]
        risk_df = risk_se.reindex(factor_se.index)[self._industry_styles + ['SIZE']]
        risk_df.dropna(inplace=True)
        factor_se = factor_se.loc[risk_df.index][factor_name]
        returns = forward_returns.loc[risk_df.index]
        
        factor_se = factor_se.groupby(level='trade_date').apply(lambda x: self.winsorize(x, method=method,
                                                                    limits=(up_limit, down_limit)))
        # factor_se = factor_se.groupby(level='trade_date').apply(lambda x: self.neutralize(x, risk_df)) # index 多了一项
        grouped = factor_se.groupby(level='trade_date')
        res = []
        for trade_date, g in grouped:
            neu = self.neutralize(g, risk_df)
            res.append(neu)
        factor_se = pd.concat(res, axis=0)

        factor_se = factor_se.groupby(level='trade_date').apply(lambda x: self.standardize(x))
        
        factor_ret, pnl, turnover_se, lc, sc = self.returns(factor_se, forward_returns['ret'], init_capital=init_capital)
        
        #计算指标
        ir = factor_ret.mean()/factor_ret.std()
        sharpe = np.sqrt(252/horizon)*ir
        turnover = turnover_se.mean()
        returns = factor_ret.sum()*252/horizon/len(factor_ret)
        fitness = sharpe * np.sqrt(abs(returns)/turnover)
        margin = factor_ret.sum()/turnover_se.sum()
        
        capital = pnl + init_capital
        running_max = np.maximum.accumulate(capital)
        drawback_data = -((running_max - capital) / running_max)
        max_drawdown = np.min(drawback_data)
        
        return {'ir':ir, 'sharpe':sharpe, 'turnover':turnover, 'returns':returns, 
                'fitness':fitness, 'margin':margin, 'max_drawdown':max_drawdown}


# -*- coding: utf-8 -*-
import os,logging,datetime,pdb,inspect,time,six
from . singleton import Singleton

logger_dict = {'info':logging.INFO,
                'warning':logging.WARNING,
                'critical':logging.CRITICAL,
                'debug':logging.DEBUG,
                'error':logging.ERROR}


class BaseLogger(object):
    def __init__(self, logger_name, log_level='info', 
                 log_format=None, log_dir=None):
        self._init_setting(logger_name = logger_name, 
                           log_level=log_level, 
                           log_format=log_format, log_dir=log_dir)
        #self.logger = None
        
    def _init_setting(self, logger_name, log_level='info', 
                 log_format=None, log_dir=None):
        default_format = '%(asctime)s - [%(filename)s:%(lineno)d] - %(name)s - %(levelname)s - %(message)s'
        default_path = '/var/log/' if 'KD_LOG_PATH' not in os.environ else os.environ['KD_LOG_PATH']
        log_path = default_path if log_dir is None else log_dir
        log_format = default_format if log_format is None else log_format
        formatter = logging.Formatter(log_format)
        self.logger = logging.getLogger(logger_name)
        console = logging.StreamHandler()
        console.setFormatter(formatter)
        self.logger.addHandler(console)
        self.set_level(log_level)
        self._config_file(log_level=log_level, logger_name=logger_name, 
                              log_path=log_path, log_format=log_format)
            
    def _config_file(self, log_level, logger_name, log_format, 
                     log_path=None):
        dir_name = os.path.join(log_path, logger_name)
        if not os.path.exists(dir_name):
            os.makedirs(dir_name)
        filename = datetime.datetime.now().date().strftime('%Y-%m-%d') + '.log'
        print(os.path.join(dir_name, filename))
        logging.basicConfig(level=logger_dict[log_level.lower()],
                           format=log_format,
                           datefmt='%m-%d %H:%M',
                           filename=os.path.join(dir_name, filename),
                           filemode='a')
        
    def reset(self, logger_name, log_level='info', 
                 log_format=None, log_dir=None):
        self._init_setting(logger_name = logger_name, 
                           log_level=log_level, 
                           log_format=log_format, log_dir=log_dir)
        return self
    
    def handle(self):
        return self.logger
    
    def set_level(self, log_level):
        self.logger.setLevel(logger_dict[log_level.lower()])
    
    def info(self, msg):
        self.logger.info(msg)
    
    def warning(self, msg):
        self.logger.warning(msg)
    
    def critical(self, msg):
        self.logger.critical(msg)
        
    def debug(self, msg):
        self.logger.debug(msg)
        
    def error(self, msg):
        self.logger.error(msg)
        
        
class RDLogger(object):
    def __init__(self, logger_name, log_level='info'):
        self._logger_name = logger_name
        self.logger = logging.getLogger(logger_name)
        console = logging.StreamHandler()
        self.logger.addHandler(console)
        self.set_level(log_level)
        
    def set_level(self, log_level):
        self.logger.setLevel(logger_dict[log_level.lower()])
        
    def message(self, level, message):
        _,file_name,line_no,function_name,_,_ = inspect.stack()[2]
        return "[%s line:%s] %s - %s - %s - %s" %(file_name, line_no,
                    time.strftime('%Y-%m-%d %H:%M:%S', time.localtime()),
                    self._logger_name, level.upper(), message)
    
    def info(self, msg):
        self.logger.info(self.message('info', msg))
    
    def warning(self, msg):
        self.logger.info(self.message('warning', msg))
    
    def critical(self, msg):
        self.logger.info(self.message('critical', msg))
        
    def debug(self, msg):
        self.logger.info(self.message('debug', msg))
        
    def error(self, msg):
        self.logger.info(self.message('error', msg))
        
        
class LoggerFactory(object):
    @classmethod
    def create_engine(self, name='custom'):
        if name == 'rd':
            return RDLogger
        else:
            return BaseLogger(name)

@six.add_metaclass(Singleton)
class CustomLogger():
    def __init__(self, name=None):
        self.logger = LoggerFactory.create_engine(name=name)
    
    def reset(self, name):
        self.logger = self.logger.reset(logger_name=name)
        global mlc_logger
        mlc_logger = self.handle()
        
    def handle(self):
        return self.logger.handle()
        
kd_logger = CustomLogger(
    'ultron' if 'KD_LOG_NAME' not in os.environ else os.environ['ULTRON_LOG_NAME']).handle()


# -*- coding: utf-8 -*-

from ... utilities.singleton import Singleton
from ultron.sentry.Analysis.SecurityValueHolders import SecurityLatestValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityCurrentValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityDiffValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySignValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityExpValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityLogValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySqrtValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityAbsValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityNormInvValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityCeilValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityFloorValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityRoundValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecuritySigmoidValueHolder
from ultron.sentry.Analysis.TechnicalAnalysis.StatelessTechnicalAnalysers import SecurityTanhValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSRankedSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSZScoreSecurityValueHolder
from ultron.sentry.Analysis.CrossSectionValueHolders import CSPercentileSecurityValueHolder


from ultron.sentry.Analysis.CrossSectionValueHolders import CSResidueSecurityValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityAddedValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecuritySubbedValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityMultipliedValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityDividedValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityLtOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityLeOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityGtOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityGeOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityEqOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityNeOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityAndOperatorValueHolder
from ultron.sentry.Analysis.SecurityValueHolders import SecurityOrOperatorValueHolder

import six,pdb

@six.add_metaclass(Singleton)
class Accumulators(object):
    def __init__(self):
        self._accumulators_pool = {
                1:SecurityCurrentValueHolder,2:SecurityDiffValueHolder,
                3:SecuritySignValueHolder,4:SecurityExpValueHolder,
                5:SecurityLogValueHolder,6:SecuritySqrtValueHolder,
                7:SecurityAbsValueHolder,8:SecurityNormInvValueHolder,
                9:SecurityCeilValueHolder,10:SecurityFloorValueHolder,
                11:SecurityRoundValueHolder,12:SecurityRoundValueHolder,
                13:CSRankedSecurityValueHolder,
                14:CSZScoreSecurityValueHolder,15:CSPercentileSecurityValueHolder,
                16:SecuritySigmoidValueHolder,17:SecurityTanhValueHolder
        }
        
        self._mutated_pool = {
                1:SecurityCurrentValueHolder,2:SecurityDiffValueHolder,
                3:SecuritySignValueHolder,4:SecurityExpValueHolder,
                5:SecurityLogValueHolder,6:SecuritySqrtValueHolder,
                7:SecurityAbsValueHolder,8:SecurityNormInvValueHolder,
                9:SecurityCeilValueHolder,10:SecurityFloorValueHolder,
                11:SecurityRoundValueHolder,12:SecurityRoundValueHolder,
                13:CSRankedSecurityValueHolder,
                14:CSZScoreSecurityValueHolder,15:CSPercentileSecurityValueHolder,
                16:SecuritySigmoidValueHolder,17:SecurityTanhValueHolder
        }
        self._cross_pool = {
                1:CSResidueSecurityValueHolder,2:SecurityAddedValueHolder,
                3:SecuritySubbedValueHolder,4:SecurityMultipliedValueHolder,
                5:SecurityDividedValueHolder,6:SecurityLtOperatorValueHolder,
                7:SecurityLeOperatorValueHolder,8:SecurityGtOperatorValueHolder,
                9:SecurityGeOperatorValueHolder,10:SecurityEqOperatorValueHolder,
                11:SecurityNeOperatorValueHolder,12:SecurityAndOperatorValueHolder,
                13:SecurityOrOperatorValueHolder
        }
        
    def transform(self, expression, is_formula=False):
        var_group  = expression.split('c_')
        formula = ''
        is_acc = False
        for i in range(len(var_group)):
            if i == 0:
                formula = '\'' + var_group[i] + '\''
            elif int(var_group[i]) > 1:
                is_acc = True
                formula = self._accumulators_pool[int(var_group[i])].__name__ + '(' + formula + ')'
        if not is_acc : formula = self._accumulators_pool[1].__name__ + '(' + formula + ')'
        return eval(formula) if is_formula else formula
    
    def dependency(self, expression_sets):
        result = {}
        expression_list = list(expression_sets)
        for expression in expression_list:
            result[expression] = ','.join(eval(expression)._dependency)
        return result
            
        
    def calc_new_factor(self, mutated_cross_columns, factor_data):
        new_factors = None
        for columns in mutated_cross_columns:
            sub_data = eval(columns).transform(factor_data, name=str(columns),
                                                category_field='code', dropna=False)
            sub_data = sub_data.reset_index().set_index('trade_date','code')
            if new_factors is None:
                new_factors = sub_data.copy()
            else:
                new_factors[str(columns)] = sub_data[str(columns)].values
        return new_factors.reset_index()[mutated_cross_columns]
            
            
    def fetch_accumulators_pool(self):
        return self._accumulators_pool
    
    def fetch_mutated_pool(self):
        return self._mutated_pool
    
    def fetch_cross_pool(self):
        return self._cross_pool
    
    def get_accumulators_pool(self, index):
        return self._accumulators_pool[index]
    
accumulators_pool = Accumulators().fetch_accumulators_pool()
mutated_pool = Accumulators().fetch_mutated_pool()
cross_pool = Accumulators().fetch_cross_pool()
transform = Accumulators().transform
dependency = Accumulators().dependency
calc_new_factor = Accumulators().calc_new_factor


# -*- coding: utf-8 -*-


cimport cython
import numpy as np
cimport numpy as np
from ultron.sentry.Utilities.Asserts cimport pyFinAssert

@cython.boundscheck(False)
@cython.wraparound(False)
cpdef to_dict(total_index, list total_category, double[:, :] matrix_values, list columns):

    cdef long start = 0
    cdef int i
    cdef long end
    cdef int k
    cdef int j

    cdef dict current_dict = {}
    cdef list index_diff_loc = list(np.where(np.diff(total_index))[0])
    cdef size_t index_diff_length = len(index_diff_loc)
    cdef list splited_values = [None] * (index_diff_length + 1)
    cdef list splited_category = [None] * (index_diff_length + 1)
    cdef size_t column_length = len(columns)

    for i in range(index_diff_length):
        end = index_diff_loc[i]
        splited_category[i] = total_category[start:end+1]
        current_dict = {}
        for j in range(start, end+1):
            current_dict[total_category[j]] = {columns[k]: matrix_values[j, k] for k in range(column_length)}
        splited_values[i] = current_dict
        start = end + 1

        pyFinAssert(len(current_dict) == len(splited_category[i]),
                    ValueError,
                    "There is duplicated category value in the snapshot ({0})".format(total_index[start]))

    splited_category[index_diff_length] = total_category[start:]
    current_dict = {}
    for j in range(start, len(total_category)):
        current_dict[total_category[j]] = {columns[k]: matrix_values[j, k] for k in range(column_length)}
    splited_values[index_diff_length] = current_dict

    return splited_category, splited_values


# -*- coding: utf-8 -*-

import numpy as np
from sklearn.linear_model import LinearRegression as LinearRegressionImpl
from sklearn.linear_model import Lasso
from sklearn.linear_model import LogisticRegression as LogisticRegressionImpl
from ultron.optimize.model.modelbase import create_model_base


class ConstLinearModelImpl(object):

    def __init__(self, weights: np.ndarray = None):
        self.weights = weights.flatten()

    def fit(self, x: np.ndarray, y: np.ndarray):
        raise NotImplementedError("Const linear model doesn't offer fit methodology")

    def predict(self, x: np.ndarray):
        return x @ self.weights

    def score(self, x: np.ndarray, y: np.ndarray) -> float:
        y_hat = self.predict(x)
        y_bar = y.mean()
        ssto = ((y - y_bar) ** 2).sum()
        sse = ((y - y_hat) ** 2).sum()
        return 1. - sse / ssto


class ConstLinearModel(create_model_base()):

    def __init__(self,
                 features=None,
                 weights: dict = None,
                 fit_target=None):
        super().__init__(features=features, fit_target=fit_target)
        if features is not None and weights is not None:
            assert(len(features) == len(weights),
                        ValueError,
                        "length of features is not equal to length of weights")
        if weights:
            self.impl = ConstLinearModelImpl(np.array([weights[name] for name in self.features]))

    def save(self):
        model_desc = super().save()
        model_desc['weight'] = list(self.impl.weights)
        return model_desc

    @classmethod
    def load(cls, model_desc: dict):
        return super().load(model_desc)

    @property
    def weights(self):
        return self.impl.weights.tolist()


class LinearRegression(create_model_base('sklearn')):

    def __init__(self, features=None, fit_intercept: bool = False, fit_target=None, **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = LinearRegressionImpl(fit_intercept=fit_intercept, **kwargs)

    def save(self) -> dict:
        model_desc = super().save()
        model_desc['weight'] = self.impl.coef_.tolist()
        return model_desc

    @property
    def weights(self):
        return self.impl.coef_.tolist()


class LassoRegression(create_model_base('sklearn')):

    def __init__(self, alpha=0.01, features=None, fit_intercept: bool = False, fit_target=None, **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = Lasso(alpha=alpha, fit_intercept=fit_intercept, **kwargs)

    def save(self) -> dict:
        model_desc = super().save()
        model_desc['weight'] = self.impl.coef_.tolist()
        return model_desc

    @property
    def weights(self):
        return self.impl.coef_.tolist()


class LogisticRegression(create_model_base('sklearn')):

    def __init__(self, features=None, fit_intercept: bool = False, fit_target=None, **kwargs):
        super().__init__(features=features, fit_target=fit_target)
        self.impl = LogisticRegressionImpl(fit_intercept=fit_intercept, **kwargs)

    def save(self) -> dict:
        model_desc = super().save()
        model_desc['weight'] = self.impl.coef_.tolist()
        return model_desc

    @property
    def weights(self):
        return self.impl.coef_.tolist()


# -*- coding: utf-8 -*-

cimport cython
from libc.math cimport exp


cdef double _M_SQRT_2 = 0.7071067811865475244008443621048490392848359376887
cdef double _M_1_SQRTPI = 0.564189583547756286948


@cython.cdivision(True)
cdef inline double pdf(double x, double denorminator, double normalizer) nogil:
        cdef double exponent = -(x * x) / denorminator
        return 0.0 if exponent <= -690.0 else normalizer * exp(exponent)


cdef double cdf(double z) nogil


cdef inline double cdf_derivative(double z) nogil:
    return pdf(z, 2.0, _M_SQRT_2 * _M_1_SQRTPI)


# -*- coding: utf-8 -*-

from ultron.tradingday.Enums._BizDayConventions cimport BizDayConventions as bdc

cpdef enum BizDayConventions:
    Following = bdc.Following
    ModifiedFollowing = bdc.ModifiedFollowing
    Preceding = bdc.Preceding
    ModifiedPreceding = bdc.ModifiedPreceding
    Unadjusted = bdc.Unadjusted
    HalfMonthModifiedFollowing = bdc.HalfMonthModifiedFollowing
    Nearest = bdc.Nearest


# -*- coding: utf-8 -*-


cimport cython
from libc.math cimport fabs
from ultron.sentry.Math.ErrorFunction cimport errorFunction
from ultron.sentry.Math.MathConstants cimport M_SQRT_2
from ultron.sentry.Math.MathConstants cimport M_1_SQRTPI
from ultron.sentry.Math.MathConstants cimport QL_EPSILON


cdef double _M_SQRT_2 = M_SQRT_2
cdef double _M_1_SQRTPI = M_1_SQRTPI
cdef double _QL_EPSILON = QL_EPSILON


@cython.cdivision(True)
cdef double cdf(double z) nogil:
    cdef double sumRes
    cdef double zsqr
    cdef double i
    cdef double g
    cdef double a
    cdef double lasta
    cdef double x
    cdef double y
    cdef double result

    result = 0.5 * (1.0 + errorFunction(z * _M_SQRT_2))

    if result <= 1.0e-8:
        sumRes = 1.0
        zsqr = z * z
        i = 1.0
        g = 1.0
        a = 2.20e308
        while True:
            lasta = a
            x = (4.0 * i - 3.0) / zsqr
            y = x * ((4.0 * i - 1.0) / zsqr)
            a = g * (x - y)
            sumRes -= a
            g *= y
            i += 1.0
            if a < 0.0:
                a = -a

            if lasta <= a or a < fabs(sumRes * _QL_EPSILON):
                break
        return -pdf(z, 2.0, _M_SQRT_2 * _M_1_SQRTPI) / z * sumRes
    return result





from .stats import (
    aggregate_returns,
    annual_return,
    annual_volatility,
    cagr,
    calmar_ratio,
    cum_returns,
    cum_returns_final,
    downside_risk,
    excess_sharpe,
    max_drawdown,
    omega_ratio,
    roll_annual_volatility,
    roll_max_drawdown,
    roll_sharpe_ratio,
    roll_sortino_ratio,
    sharpe_ratio,
    simple_returns,
    sortino_ratio,
    tail_ratio,
    _adjust_returns
)

from .periods import (
    DAILY,
    WEEKLY,
    MONTHLY,
    QUARTERLY,
    YEARLY
)


# -*- coding: utf-8 -*-

cimport numpy as np

cdef class SeriesValues(object):

    cdef public dict name_mapping
    cdef public np.ndarray values
    cdef public np.ndarray name_array

    cpdef SeriesValues mask(self, np.ndarray flags)
    cpdef list index(self)
    cpdef SeriesValues rank(self, SeriesValues groups=*)
    cpdef SeriesValues top_n(self, int n, SeriesValues groups=*)
    cpdef SeriesValues bottom_n(self, int n, SeriesValues groups=*)
    cpdef SeriesValues top_n_percentile(self, double n, SeriesValues groups=*)
    cpdef SeriesValues bottom_n_percentile(self, double n, SeriesValues groups=*)
    cpdef SeriesValues zscore(self, SeriesValues groups=*)
    cpdef SeriesValues unit(self)

    cpdef SeriesValues mean(self, SeriesValues groups=*)
    cpdef SeriesValues fillna(self, SeriesValues groups=*)
    cpdef SeriesValues percentile(self, SeriesValues groups=*)
    cpdef double dot(self, SeriesValues right)
    cpdef SeriesValues res(self, SeriesValues right)
    cpdef dict to_dict(self)


cdef SeriesValues residue(SeriesValues left, SeriesValues right)
cpdef SeriesValues s_maximum(SeriesValues left, SeriesValues right)
cpdef SeriesValues s_minimum(SeriesValues left, SeriesValues right)



# -*- coding: utf-8 -*-

from . combine_method import equal_combine, hist_ret_combine, hist_ic_combine, max_icir_combine, max_ic_combine

func_dict = {'equal_combine':equal_combine,
                'hist_ret_combine':hist_ret_combine,
                'hist_ic_combine':hist_ic_combine,
                'max_icir_combine':max_icir_combine,
                'max_ic_combine':max_ic_combine}
    
    
class CombineEngine(object):
    @classmethod
    def create_engine(cls, ce_name):
        return func_dict[ce_name]


# -*- coding: utf-8 -*-


#import ultron.sentry.api as api

#import ultron.sentry.Math as Math
#import ultron.sentry.Analysis as Analysis
#import ultron.sentry.Utilities as Utilities

__all__ = ['__version__']

__version__ = "0.0.1"


# -*- coding: utf-8 -*-

import copy
import pandas as pd
from ultron.sentry.Analysis.SecurityValueHolders import SecurityValueHolder
from ultron.sentry.api import transform as transform_impl


def factor_translator(factor_pool):

    if not factor_pool:
        return None, None

    if isinstance(factor_pool, str):
        return {factor_pool: factor_pool}, [factor_pool]
    elif isinstance(factor_pool, SecurityValueHolder):
        return {str(factor_pool): factor_pool}, sorted(factor_pool.fields)
    elif isinstance(factor_pool, dict):
        dependency = set()
        for k, v in factor_pool.items():
            pyFinAssert(isinstance(k, str), ValueError, 'factor_name {0} should be string.'.format(k))
            pyFinAssert(isinstance(v, SecurityValueHolder) or isinstance(v, str),
                        ValueError,
                        'expression {0} should be a value hodler or a string.'.format(v))

            if isinstance(v, str):
                dependency = dependency.union([v])
            else:
                dependency = dependency.union(v.fields)
        return factor_pool, sorted(dependency)
    elif isinstance(factor_pool, list):
        factor_dict = {}
        dependency = set()
        k = 1
        for i, f in enumerate(factor_pool):
            if isinstance(f, str):
                factor_dict[f] = f
                dependency = dependency.union([f])
            elif isinstance(f, SecurityValueHolder):
                factor_dict[str(f)] = f
                dependency = dependency.union(f.fields)
                k += 1
        return factor_dict, sorted(dependency)
    else:
        raise ValueError('{0} is not in valid format as factors'.format(factor_pool))


class Transformer(object):

    def __init__(self,
                 expressions):
        expression_dict, expression_dependency = \
            factor_translator(copy.deepcopy(expressions))

        if expression_dict:
            self.names = sorted(expression_dict.keys())
            self.expressions = [expression_dict[n] for n in self.names]
            self.dependency = expression_dependency
        else:
            self.names = []
            self.expressions = []
            self.dependency = []

    def transform(self, group_name, data):
        if len(data) > 0:
            transformed_data = transform_impl(data,
                                              self.expressions,
                                              self.names,
                                              group_name,
                                              dropna=False)
            return transformed_data
        else:
            return pd.DataFrame()


import numpy as np
import numba as nb



@nb.njit(nogil=True, cache=True)
def simple_abssum(x, axis=0):
    length, width = x.shape

    if axis == 0:
        res = np.zeros(width)
        for i in range(length):
            for j in range(width):
                res[j] += abs(x[i, j])

    elif axis == 1:
        res = np.zeros(length)
        for i in range(length):
            for j in range(width):
                res[i] += abs(x[i, j])
    else:
        raise ValueError("axis value is not supported")
    return res


@nb.njit(nogil=True, cache=True)
def group_mapping(groups: np.ndarray) -> np.ndarray:
    length = groups.shape[0]
    order = groups.argsort()
    res = np.zeros(length, dtype=order.dtype)

    start = 0
    res[order[0]] = start
    previous = groups[order[0]]

    for i in range(1, length):
        curr_idx = order[i]
        curr_val = groups[curr_idx]
        if curr_val != previous:
            start += 1
            res[curr_idx] = start
        else:
            res[curr_idx] = start
        previous = curr_val
    return res


@nb.njit(nogil=True, cache=True)
def agg_std(groups, x, ddof=1):
    max_g = groups.max()
    length, width = x.shape
    res = np.zeros((max_g + 1, width), dtype=np.float64)
    sumsq = np.zeros((max_g + 1, width), dtype=np.float64)
    bin_count = np.zeros(max_g + 1, dtype=np.int32)

    for i in range(length):
        for j in range(width):
            res[groups[i], j] += x[i, j]
            sumsq[groups[i], j] += x[i, j] * x[i, j]
        bin_count[groups[i]] += 1

    for i in range(max_g + 1):
        curr = bin_count[i]
        for j in range(width):
            res[i, j] = math.sqrt((sumsq[i, j] - res[i, j] * res[i, j] / curr) / (curr - ddof))
    return res


@nb.njit(nogil=True, cache=True)
def agg_sum(groups, x):
    max_g = groups.max()
    length, width = x.shape
    res = np.zeros((max_g + 1, width), dtype=np.float64)

    for i in range(length):
        for j in range(width):
            res[groups[i], j] += x[i, j]
    return res


@nb.njit(nogil=True, cache=True)
def agg_sqrsum(groups, x):
    max_g = groups.max()
    length, width = x.shape
    res = np.zeros((max_g + 1, width), dtype=np.float64)

    for i in range(length):
        for j in range(width):
            res[groups[i], j] += x[i, j] * x[i, j]

    res = np.sqrt(res)
    return res


@nb.njit(nogil=True, cache=True)
def agg_abssum(groups, x):
    max_g = groups.max()
    length, width = x.shape
    res = np.zeros((max_g + 1, width), dtype=np.float64)

    for i in range(length):
        for j in range(width):
            res[groups[i], j] += abs(x[i, j])
    return res


@nb.njit(nogil=True, cache=True)
def agg_mean(groups, x):
    max_g = groups.max()
    length, width = x.shape
    res = np.zeros((max_g+1, width), dtype=np.float64)
    bin_count = np.zeros(max_g+1, dtype=np.int32)

    for i in range(length):
        for j in range(width):
            res[groups[i], j] += x[i, j]
        bin_count[groups[i]] += 1

    for i in range(max_g+1):
        curr = bin_count[i]
        for j in range(width):
            res[i, j] /= curr
    return res

@nb.njit(nogil=True, cache=True)
def copy_value(groups, source):
    length = groups.shape[0]
    width = source.shape[1]
    destination = np.zeros((length, width))
    for i in range(length):
        k = groups[i]
        for j in range(width):
            destination[i, j] = source[k, j]
    return destination


@nb.njit(nogil=True, cache=True)
def scale_value(groups, source, x, scale):
    length, width = x.shape
    destination = x.copy()
    for i in range(length):
        k = groups[i]
        for j in range(width):
            destination[i, j] /= source[k, j] / scale
    return destination


def encode(obj: object) -> str:
    encoded = base64.encodebytes(pickle.dumps(obj))
    return encoded.decode('ascii')


def decode(str_repr: str):
    encoded = str_repr.encode('ascii')
    return pickle.loads(base64.decodebytes(encoded))

def list_eq(lhs: list, rhs: list):
    if not lhs and not rhs:
        return True

    if len(lhs) != len(rhs):
        return False

    for i, v1 in enumerate(lhs):
        if v1 != rhs[i]:
            return False
    return True

def transform(groups: np.ndarray,
              x: np.ndarray,
              func: str,
              ddof: int = 1,
              scale: float = 1.) -> np.ndarray:
    if func == 'mean':
        value_data = agg_mean(groups, x)
    elif func == 'std':
        value_data = agg_std(groups, x, ddof=ddof)
    elif func == 'sum':
        value_data = agg_sum(groups, x)
    elif func == 'abssum' or func == 'scale':
        value_data = agg_abssum(groups, x)
    elif func == 'sqrsum' or func == 'project':
        value_data = agg_sqrsum(groups, x)
    else:
        raise ValueError('({0}) is not recognized as valid functor'.format(func))

    if func == 'scale' or func == 'project':
        return scale_value(groups, value_data, x, scale)
    else:
        return copy_value(groups, value_data)



# -*- coding: utf-8 -*-

cimport numpy as np
from ultron.sentry.Math.Accumulators.impl cimport Deque
from ultron.sentry.Math.Accumulators.IAccumulators cimport Accumulator


cdef class StatefulValueHolder(Accumulator):

    cdef public Deque _deque

    cpdef size_t size(self)
    cpdef bint isFull(self)


cdef class Shift(StatefulValueHolder):

    cdef public double _popout
    cdef public Accumulator _x

    cpdef int lag(self)
    cpdef push(self, dict data)
    cpdef double result(self)


cdef class Delta(StatefulValueHolder):

    cdef public double _popout
    cdef double _current
    cdef public Accumulator _x

    cpdef int lag(self)
    cpdef push(self, dict data)
    cpdef double result(self)


cdef class SingleValuedValueHolder(StatefulValueHolder):

    cdef public Accumulator _x


cdef class SortedValueHolder(SingleValuedValueHolder):

    cdef list _sortedArray
    cdef double _cur_pos
    cpdef push(self, dict data)

cdef class MovingMax(SortedValueHolder):

    cpdef double result(self)


cdef class MovingArgMax(SortedValueHolder):

    cpdef double result(self)


cdef class MovingMin(SortedValueHolder):

    cpdef double result(self)


cdef class MovingArgMin(SortedValueHolder):

    cpdef double result(self)


cdef class MovingRank(SortedValueHolder):

    cpdef double result(self)


cdef class MovingQuantile(SortedValueHolder):

    cpdef double result(self)


cdef class MovingAllTrue(SingleValuedValueHolder):

    cdef public size_t _countedTrue

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingAnyTrue(SingleValuedValueHolder):

    cdef public size_t _countedTrue

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingSum(SingleValuedValueHolder):

    cdef public double _runningSum

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingAverage(SingleValuedValueHolder):

    cdef public double _runningSum

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingDecay(SingleValuedValueHolder):
    cdef public double _runningSum
    cdef double _runningWeightedSum
    cdef double _newestValue

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingPositiveAverage(SingleValuedValueHolder):

    cdef public double _runningPositiveSum
    cdef public int _runningPositiveCount

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingPositiveDifferenceAverage(SingleValuedValueHolder):

    cdef public MovingAverage _runningAverage

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingNegativeDifferenceAverage(SingleValuedValueHolder):

    cdef public MovingAverage _runningAverage

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingRSI(SingleValuedValueHolder):

    cdef public MovingPositiveDifferenceAverage _posDiffAvg
    cdef public MovingNegativeDifferenceAverage _negDiffAvg

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingNegativeAverage(SingleValuedValueHolder):

    cdef public double _runningNegativeSum
    cdef public int _runningNegativeCount

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingVariance(SingleValuedValueHolder):

    cdef public double _runningSum
    cdef public double _runningSumSquare
    cdef public int _isPop

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingStandardDeviation(SingleValuedValueHolder):

    cdef public double _runningSum
    cdef public double _runningSumSquare
    cdef public int _isPop

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingNegativeVariance(SingleValuedValueHolder):

    cdef public double _runningNegativeSum
    cdef public double _runningNegativeSumSquare
    cdef public int _runningNegativeCount
    cdef public int _isPop

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingCountedPositive(SingleValuedValueHolder):

    cdef public int _counts

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingCountedNegative(SingleValuedValueHolder):

    cdef public int _counts

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingCorrelation(StatefulValueHolder):

    cdef public double _runningSumLeft
    cdef public double _runningSumRight
    cdef public double _runningSumSquareLeft
    cdef public double _runningSumSquareRight
    cdef public double _runningSumCrossSquare
    cdef Accumulator _x
    cdef Accumulator _y
    cdef Deque _deque_y

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingProduct(SingleValuedValueHolder):

    cdef public double _runningProduct

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MACD(Accumulator):

    cdef public Accumulator _short_average
    cdef public Accumulator _long_average

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingLogReturn(SingleValuedValueHolder):

    cdef public double _runningReturn

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingSharp(StatefulValueHolder):

    cdef public MovingAverage _mean
    cdef public MovingVariance _var
    cdef Accumulator _x
    cdef Accumulator _y
    cdef Deque _deque_y

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingSortino(StatefulValueHolder):

    cdef public MovingAverage _mean
    cdef public MovingNegativeVariance _negativeVar
    cdef Accumulator _x
    cdef Accumulator _y
    cdef Deque _deque_y

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingDrawdown(StatefulValueHolder):

    cdef MovingMax _maxer
    cdef double _runningCum
    cdef double _currentMax
    cdef Accumulator _x

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingMaxDrawdown(StatefulValueHolder):

    cdef MovingDrawdown _drawdownCalculator
    cdef MovingMin _minimer
    cdef Accumulator _x

    cpdef push(self, dict data)
    cpdef double result(self)


cdef class MovingResidue(StatefulValueHolder):

    cdef public double _cross
    cdef public double _xsquare
    cdef public double _lastx
    cdef public double _lasty
    cdef Accumulator _x
    cdef Accumulator _y
    cdef Deque _deque_y

    cpdef push(self, dict data)
    cpdef double result(self)
    cpdef bint isFull(self)
